#include "EventManager.h"
#include <algorithm>
#include <iostream>
#include "property/MyPropertySystem.h"

const float FRAMETIME = 0.1f;

EventManager::EventManager()
    : m_remainedTime(0.0f), m_accumulatedTime(0.0f)
{

}

EventManager::~EventManager()
{
    EventTicketList::iterator iter = m_ticketList.begin();
    for(;iter != m_ticketList.end(); ++iter)
    {
        EventTicket& ticket = *iter;
        if(ticket.pProperty)
            delete ticket.pProperty;
    }
    m_ticketList.clear();
}

bool EventManager::AddEvent(EventId eventId, EventFunc eventFunc)
{
    return AddEvent(eventId, eventFunc, eventId);
}

bool EventManager::AddEvent(EventId eventId, EventFunc eventFunc, EntryName entryName)
{
    EventHandler handler = { eventId, eventFunc, entryName};

    auto iter = m_eventMap.find(eventId);
    if (iter != m_eventMap.end())
    {
        EventList& list = iter->second;

        auto it = std::find_if(list.begin(), list.end(), [&entryName](EventHandler const& item) -> bool {
            return item.entryName.compare(entryName) == 0;
        });

        if (it != list.end())
        {
            return false;
        }
        list.push_back(handler);
        m_entryMap[entryName] = eventId;
        return true;
    }

    EventList list;
    list.push_back(handler);
    m_eventMap[eventId] = list;
    m_entryMap[entryName] = eventId;
    return true;
}

bool EventManager::RemoveEvent(EntryName entryName)
{
    auto idIter = m_entryMap.find(entryName);
    if (idIter == m_entryMap.end())
        return false;

    auto iter = m_eventMap.find(idIter->second);
    if (iter != m_eventMap.end())
    {
        EventList& list = iter->second;
        auto it = std::find_if(list.begin(), list.end(), [&entryName](EventHandler const& item) -> bool {
            return item.entryName.compare(entryName) == 0;
        });

        if (it != list.end())
        {
            list.erase(it);
            return true;
        }
    }

    return false;
}

void EventManager::SendEvent(EventId eventId, const MyPropertySystem& param)
{
    auto iter = m_eventMap.find(eventId);
    if (iter != m_eventMap.end())
    {
        EventList& list = iter->second;

        for (const EventHandler& item : list)
        {
            item.func(&param);
        }
    }
}

void EventManager::PostEvent(EventId eventId, const MyPropertySystem& param, float duration)
{
    EventTicket ticket;
    ticket.eventId = eventId;
    ticket.pProperty = new MyPropertySystem(param);
    ticket.durationTime = duration;
    ticket.startTime = m_accumulatedTime;
    m_ticketList.push_back(ticket);
}

void EventManager::Update(float deltaTime)
{
    m_remainedTime += deltaTime;
    m_accumulatedTime += deltaTime;
    if (m_remainedTime > FRAMETIME) 
    {
        m_remainedTime -= FRAMETIME;
        EventTicketList::iterator iter = m_ticketList.begin();
        while(iter != m_ticketList.end()) 
        {
            EventTicket& ticket = *iter;
            if ((m_accumulatedTime - ticket.startTime) >= ticket.durationTime)
            {
                SendEvent(ticket.eventId, *(ticket.pProperty));
                if(ticket.pProperty)
                    delete ticket.pProperty;
                m_ticketList.erase(iter++);
            }
            else
                ++iter;
        }
    }
}